@@ -33,23 +33,20 @@ class Mlx9064x:
         else:
             self.hw = hw
         self.i2c_addr = i2c_addr
-        self.frame_rate = property(self.get_frame_rate, self.set_frame_rate)
         self.calc_params = TCalcParams()
         self.m_lFilterTgcDepth = 8
         self.m_arrLastTgc = [[0] * TCalcParams.NUM_PAGES] * TCalcParams.NUM_TGC
-        self.m_fEmissivity = 1.0
+        self.emissivity = 1.0
 
         self.hw.connect()
-        self.set_frame_rate(frame_rate)
+        self.frame_rate = frame_rate
         if self.hw.get_sensor_type(0x33) == 0:
             self.frame_length_bytes = 32 * 26 * 2
             self.eeprom = Mlx90640EEPROM(self)
         else:
-            self.frame_length_bytes = 16 * 16 * 2
             self.frame_length_bytes = 16 * 16 * 2
             self.eeprom = Mlx90641EEPROM(self)
 
-
     def init(self):
         self.eeprom.read_eeprom_from_device()
         self.calculate_parameters()
@@ -60,17 +57,15 @@ class Mlx9064x:
         if callable(getattr(self.hw, 'set_vdd', None)):
             self.hw.set_vdd(vdd)
 
-    def get_frame_rate(self):
-        return self.frame_rate
+    @property
+    def frame_rate(self):
+        return self.__frame_rate
 
-    def clear_error(self, frame_rate_hz):
-        if callable(getattr(self.hw, 'clear_error', None)):
-            self.hw.clear_error(self.i2c_addr, frame_rate_hz)
-
-    def set_frame_rate(self, frame_rate_hz):
+    @frame_rate.setter
+    def frame_rate(self, frame_rate):
         """
         Set the refresh rate of the camera
-        :param float frame_rate_hz: the new frame rate for the camera
+        :param float frame_rate: the new frame rate for the camera
         """
         # set the refresh rate on the chip
         ctrl_reg_1, status = self.hw.i2c_read(self.i2c_addr, 0x800D, 2)
@@ -78,10 +73,10 @@ class Mlx9064x:
             raise I2CAcknowledgeError("Error during read of Control register 1")
         ctrl_reg_1 = struct.unpack(">H", ctrl_reg_1)[0]  # TODO find endian
 
-        frame_rate_code = Mlx9064x.frame_rate_to_bit_mask(frame_rate_hz)
+        frame_rate_code = Mlx9064x.frame_rate_to_bit_mask(frame_rate)
         if frame_rate_code is None:
             raise ValueError("Invalid value for frame rate: {}; valid values are {}"
-                             .format(frame_rate_hz, [0.5, 1, 2, 4, 8, 16, 32, 64]))
+                             .format(frame_rate, [0.5, 1, 2, 4, 8, 16, 32, 64]))
         else:
             ctrl_reg_1 &= 0xFC7F  # clear the 3 bits that represent the frame rate
             ctrl_reg_1 |= frame_rate_code << 7
@@ -89,10 +84,15 @@ class Mlx9064x:
         status = self.hw.i2c_write(self.i2c_addr, 0x800D, struct.pack(">H", ctrl_reg_1))
         if 0 != status:
             raise I2CAcknowledgeError("Error during write of Control register 1")
+        self.__frame_rate = frame_rate
 
         self.set_vdd(3.3)
         if self.hw.support_buffer:
-            self.hw.start_data_acquisition(self.i2c_addr, frame_rate_hz)
+            self.hw.start_data_acquisition(self.i2c_addr, frame_rate)
+
+    def clear_error(self, frame_rate_hz):
+        if callable(getattr(self.hw, 'clear_error', None)):
+            self.hw.clear_error(self.i2c_addr, frame_rate_hz)
 
     @staticmethod
     def frame_rate_to_bit_mask(f_rate):
@@ -150,10 +150,11 @@ class Mlx9064x:
     def i2c_write(self, addr, data):
         return self.hw.i2c_write(self.i2c_addr, addr, data)
 
-    def do_compensation(self, raw_frame):
+    def do_compensation(self, raw_frame, add_ambient_temperature=False):
         """
         Calculates the temperatures for each pixel
         :param raw_frame: the raw frame
+        :param add_ambient_temperature: flag to add ambient temperature at the end of the frame array.
         :return: the calculated frame as a one dimensional array
         """
         if self.hw.sensor_type == 0:
@@ -310,19 +311,24 @@ class Mlx9064x:
             Pix_GainComp = raw_frame[i] * dGainComp
 
             # 2. Pixel offset compensation
+            Pix_os_ref = 0
+            Kta = 0
+            Kv = 0
             if self.hw.sensor_type == 0:
-                Pix_os = self.calc_params.Pix_os_ref[tidx][idxGlobal] * \
-                (1 + self.calc_params.Kta[tidx][idxGlobal] * dDeltaTa) * \
-                (1 + self.calc_params.Kv[tidx][idxGlobal] * dDeltaV)
+                Pix_os_ref = self.calc_params.Pix_os_ref[tidx][idxGlobal]
+                Kta = self.calc_params.Kta[tidx][idxGlobal]
+                Kv = self.calc_params.Kv[tidx][idxGlobal]
             else:
                 if page:
-                    Pix_os = self.calc_params.Pix_os_ref_SP1[tidx][idxGlobal] * \
-                    (1 + self.calc_params.Kta[tidx][idxGlobal] * dDeltaTa) * \
-                    (1 + self.calc_params.Kv[tidx][idxGlobal] * dDeltaV)
+                    Pix_os_ref = self.calc_params.Pix_os_ref_SP1[tidx][idxGlobal]
+                    Kta = self.calc_params.Kta[tidx][idxGlobal]
+                    Kv = self.calc_params.Kv[tidx][idxGlobal]
                 else:
-                    Pix_os = self.calc_params.Pix_os_ref_SP0[tidx][idxGlobal] * \
-                    (1 + self.calc_params.Kta[tidx][idxGlobal] * dDeltaTa) * \
-                    (1 + self.calc_params.Kv[tidx][idxGlobal] * dDeltaV)
+                    Pix_os_ref = self.calc_params.Pix_os_ref_SP0[tidx][idxGlobal]
+                    Kta = self.calc_params.Kta[tidx][idxGlobal]
+                    Kv = self.calc_params.Kv[tidx][idxGlobal]
+
+            Pix_os = Pix_os_ref * (1 + Kta * dDeltaTa) * (1 + Kv * dDeltaV)
 
             # 3. calculating offset free IR data
             Pix_comp = Pix_GainComp - Pix_os
@@ -350,9 +356,96 @@ class Mlx9064x:
                         To = To1
 
             result_frame[i] = To
+        if add_ambient_temperature:
+            result_frame.append(Tamb)
         return result_frame
 
-    def set_m_fEmissivity(self, emissivity):
+    def do_handle_bad_pixels(self, compensated_frame):
+        result = compensated_frame
+        if self.hw.sensor_type == 0:  # 90640
+            for pixel_i in self.eeprom.bad_pixels:
+                col = pixel_i & 0x1F
+                row = pixel_i >> 5
+                if pixel_i == 0 << 5 + 0:  # upper left corner
+                    result[pixel_i] = (result[0 << 5 + 1] + result[1 << 5 + 0] + result[1 << 5 + 1]) / 3
+                    continue
+                if pixel_i == 0 << 5 + 31:  # upper right corner
+                    result[pixel_i] = (result[0 << 5 + 30] + result[1 << 5 + 31] + result[1 << 5 + 30]) / 3
+                    continue
+                if pixel_i == 23 << 5 + 0:  # lower left corner
+                    result[pixel_i] = (result[23 << 5 + 1] + result[22 << 5 + 0] + result[22 << 5 + 1]) / 3
+                    continue
+                if pixel_i == 23 << 5 + 31:  # lower right corner
+                    result[pixel_i] = (result[23 << 5 + 30] + result[22 << 5 + 31] + result[22 << 5 + 30]) / 3
+                    continue
+
+                if col == 0:  # first col
+                    result[pixel_i] = (result[(row - 1) << 5 + col] + result[(row + 1) << 5 + col] + result[
+                        row << 5 + col + 1]) / 3
+                    continue
+                if col == 31:  # last col
+                    result[pixel_i] = (result[(row - 1) << 5 + col] + result[(row + 1) << 5 + col] + result[
+                        row << 5 + col - 1]) / 3
+                    continue
+
+                if row == 0:  # first row
+                    result[pixel_i] = (result[row << 5 + col - 1] + result[row << 5 + col + 1] + result[
+                        (row + 1) << 5 + col]) / 3
+                    continue
+                if row == 23:  # last row
+                    result[pixel_i] = (result[row << 5 + col - 1] + result[row << 5 + col + 1] + result[
+                        (row - 1) << 5 + col]) / 3
+                    continue
+                # pixel not on border
+                result[pixel_i] = (result[row << 5 + col - 1] + result[row << 5 + col + 1] + result[
+                    (row - 1) << 5 + col] + result[(row + 1) << 5 + col]) / 4
+
+        if self.hw.sensor_type == 1:  # 90641
+            for pixel_i in self.eeprom.bad_pixels:
+                col = pixel_i & 0x0F
+                row = pixel_i >> 4
+                if pixel_i == 0 << 4 + 0:  # upper left corner
+                    result[pixel_i] = (result[0 << 4 + 1] + result[1 << 4 + 0] + result[1 << 4 + 1]) / 3
+                    continue
+                if pixel_i == 0 << 4 + 15:  # upper right corner
+                    result[pixel_i] = (result[0 << 4 + 14] + result[1 << 4 + 15] + result[1 << 4 + 14]) / 3
+                    continue
+                if pixel_i == 11 << 4 + 0:  # lower left corner
+                    result[pixel_i] = (result[11 << 4 + 1] + result[10 << 4 + 0] + result[10 << 4 + 1]) / 3
+                    continue
+                if pixel_i == 11 << 4 + 15:  # lower right corner
+                    result[pixel_i] = (result[11 << 4 + 14] + result[10 << 4 + 15] + result[10 << 4 + 14]) / 3
+                    continue
+
+                if col == 0:  # first col
+                    result[pixel_i] = (result[(row - 1) << 4 + col] + result[(row + 1) << 4 + col] + result[
+                        row << 4 + col + 1]) / 3
+                    continue
+                if col == 15:  # last col
+                    result[pixel_i] = (result[(row - 1) << 4 + col] + result[(row + 1) << 4 + col] + result[
+                        row << 4 + col - 1]) / 3
+                    continue
+
+                if row == 0:  # first row
+                    result[pixel_i] = (result[row << 4 + col - 1] + result[row << 4 + col + 1] + result[
+                        (row + 1) << 4 + col]) / 3
+                    continue
+                if row == 11:  # last row
+                    result[pixel_i] = (result[row << 4 + col - 1] + result[row << 4 + col + 1] + result[
+                        (row - 1) << 4 + col]) / 3
+                    continue
+                # pixel not on border
+                result[pixel_i] = (result[row << 4 + col - 1] + result[row << 4 + col + 1] + result[
+                    (row - 1) << 4 + col] + result[(row + 1) << 4 + col]) / 4
+
+        return result
+
+    @property
+    def emissivity(self):
+        return self.m_fEmissivity
+
+    @emissivity.setter
+    def emissivity(self, emissivity):
         self.m_fEmissivity = emissivity
 
     def calculate_parameters(self):
@@ -1005,6 +1098,9 @@ class Mlx90640EEPROM:
         self.device = device
         self.eeprom = None
         self.eeprom_size = 0x680
+        self.outlier_pixels = []
+        self.broken_pixels = []
+        self.bad_pixels = []
 
     def get_bit(self, index, lsb):
         return (self.eeprom[index] & (1 << lsb)) != 0
@@ -1044,6 +1140,9 @@ class Mlx90640EEPROM:
             for i in range(self.eeprom_size // 2):
                 first_read[i] |= consecutive_read[i]
         self.eeprom = first_read
+        self.extract_broken_pixels()
+        self.extract_outlier_pixels()
+        self.get_bad_pixels()
 
     def get_parameter_code(self, param_id: ParameterCodesEEPROM, index=None):
         """
@@ -1074,6 +1173,27 @@ class Mlx90640EEPROM:
         else:
             ValueError("invalid eeprom parameter at {}".format(param_id))
 
+    def extract_outlier_pixels(self):
+        self.outlier_pixels = []
+        for pixel_i in range(768):
+            if self.eeprom[pixel_i + 64] & 0x0001:
+                self.outlier_pixels.append(pixel_i)
+        return self.outlier_pixels
+
+    def extract_broken_pixels(self):
+        self.broken_pixels = []
+        for pixel_i in range(768):
+            if self.eeprom[pixel_i + 64] == 0:
+                self.broken_pixels.append(pixel_i)
+        return self.broken_pixels
+
+    def get_bad_pixels(self):
+        self.bad_pixels = self.broken_pixels
+        for pixel_i in self.outlier_pixels:
+            if pixel_i not in self.bad_pixels:
+                self.bad_pixels.append(pixel_i)
+        return self.bad_pixels
+
 
 class Mlx90641EEPROM:
     eeprom_map = {
@@ -1142,6 +1262,7 @@ class Mlx90641EEPROM:
         self.device = device
         self.eeprom = None
         self.eeprom_size = 0x680
+        self.bad_pixels = []
 
     def get_bit(self, index, lsb):
         return (self.eeprom[index] & (1 << lsb)) != 0
@@ -1181,6 +1302,7 @@ class Mlx90641EEPROM:
             for i in range(self.eeprom_size // 2):
                 first_read[i] |= consecutive_read[i]
         self.eeprom = first_read
+        self.get_bad_pixels()
 
     def get_parameter_code(self, param_id: ParameterCodesEEPROM, index=None):
         """
@@ -1211,6 +1333,17 @@ class Mlx90641EEPROM:
         else:
             ValueError("invalid eeprom parameter at {}".format(param_id))
 
+    def get_bad_pixels(self):
+        self.bad_pixels = []
+        for pixel_i in range(192):
+            if self.eeprom[0x0040 + pixel_i] == 0:
+                if self.eeprom[0x0100 + pixel_i] == 0:
+                    if self.eeprom[0x01C0 + pixel_i] == 0:
+                        if self.eeprom[0x0680 + pixel_i] == 0:
+                            self.bad_pixels.append(pixel_i)
+
+        return self.bad_pixels
+
 
 def main():
     # demo to test the file.. get a single frame from first comport found that has a EVB90640.
